/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */

const Atk = imports.gi.Atk;
const GMenu = imports.gi.GMenu;
const Lang = imports.lang;
const Shell = imports.gi.Shell;
const St = imports.gi.St;
const Clutter = imports.gi.Clutter;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const Gtk = imports.gi.Gtk;
const GLib = imports.gi.GLib;
const Signals = imports.signals;
const Layout = imports.ui.layout;
const Pango = imports.gi.Pango;
const Tweener = imports.ui.tweener;

const Gettext = imports.gettext.domain('gnome-shell-extensions');
const _ = Gettext.gettext;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

const appSys = Shell.AppSystem.get_default();

const APPLICATION_ICON_SIZE = 64;
const APPLICATION_BOX_WIDTH = 125;
const HORIZ_FACTOR = 5;
const NAVIGATION_REGION_OVERSHOOT = 50;

const ApplicationMenuItem = new Lang.Class({
    Name: 'ApplicationMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function(button, app) {
	this.parent();
	this._app = app;
        this._button = button;

        this._iconBin = new St.Bin();
        let appName = app.get_name();
        let line1 = appName;
        let line2 = "";
        let spacePosition = appName.indexOf(" ");
        if (spacePosition > 0) {
            line1 = appName.substr(0,spacePosition);
            line2 = appName.substr(spacePosition+1);
        }
        let appLabel1 = new St.Label({ text: line1 });
        let appLabel2 = new St.Label({ text: line2 });

        this.actor.label_actor = appLabel1;

        let textureCache = St.TextureCache.get_default();
        let iconThemeChangedId = textureCache.connect('icon-theme-changed',
                                                      Lang.bind(this, this._updateIcon));

        let container=new St.BoxLayout({ x_expand: true, vertical: true, reactive: true, y_align: St.Align.MIDDLE });
        container.add(this._iconBin, { x_align: St.Align.MIDDLE });
        container.add(appLabel1, { x_fill: false, x_expand: false, x_align: St.Align.MIDDLE });
        if (spacePosition > 0)
            container.add(appLabel2, { x_fill: false, x_expand: false, x_align: St.Align.MIDDLE });

        this.actor.add_child(container);

        this.actor.connect('destroy', Lang.bind(this,
            function() {
                textureCache.disconnect(iconThemeChangedId);
            }));
        this._updateIcon();
    },

    activate: function(event) {
	this._app.open_new_window(event.get_time());
        this._button.menu.toggle();
	this.parent(event);
    },

    setActive: function(active, params) {
        if (active)
            this._button.scrollToButton(this);
        this.parent(active, params);
    },

    _getPreferredWidth: function(actor, forHeight, alloc) {
        alloc.min_size = alloc.natural_size = -1;
    },

    _updateIcon: function() {
        this._iconBin.set_child(this._app.create_icon_texture(APPLICATION_ICON_SIZE));
    }
});

const ApplicationPageItem = new Lang.Class({
    Name: 'ApplicationPageItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function(button, page) {
	this.parent();
        this._button = button;
        this._page = page;

        let pageLabel = new St.Label({ text: ""+page });
        this.actor.add_child(pageLabel);

        this.actor.connect('destroy', Lang.bind(this,
            function() {
            }));
    },

    activate: function(event) {
        this._button.scrollToPage(this._page-1);
	this.parent(event);
    },

    setActive: function(active, params) {
        this.parent(active, params);
    },

    _getPreferredWidth: function(actor, forHeight, alloc) {
        alloc.min_size = alloc.natural_size = -1;
    },
});


const CategoryMenuItem = new Lang.Class({
    Name: 'CategoryMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function(button, category) {
	this.parent();
	this._category = category;
        this._button = button;

        let name;
        if (this._category == "*")
            name = _("All");
        else if (this._category)
            name = this._category.get_name();

        this.actor.add_child(new St.Label({ text: name }));
    },

    activate: function(event) {
        this._button.selectCategory(this._category);
    }
});

const HotCorner = new Lang.Class({
    Name: 'HotCorner',
    Extends: Layout.HotCorner,

    _onCornerEntered : function() {
        if (!this._entered) {
            this._entered = true;
            if (!Main.overview.animationInProgress) {
                this._activationTime = Date.now() / 1000;
                this.rippleAnimation();
                Main.overview.toggle();
            }
        }
        return false;
    }
});

const ApplicationsMenu = new Lang.Class({
    Name: 'ApplicationsMenu',
    Extends: PopupMenu.PopupMenu,

    _init: function(sourceActor, arrowAlignment, arrowSide, button) {
        this.parent(sourceActor, arrowAlignment, arrowSide);
        this._button = button;
    },

    isEmpty: function() {
	return false;
    },

    open: function(animate) {
        this._button.hotCorner.setBarrierSize(0);
        if (this._button.hotCorner.actor) // fallback corner
            this._button.hotCorner.actor.hide();
        this.parent(animate);
    },

    close: function(animate) {
        let size = Main.layoutManager.panelBox.height;
        this._button.hotCorner.setBarrierSize(size);
        if (this._button.hotCorner.actor) // fallback corner
            this._button.hotCorner.actor.show();
        this.parent(animate);
    },

    toggle: function() {
        if (!this.isOpen) {
            this._button.selectCategory("*");
            if (Main.overview.visible)
                Main.overview.hide();
        }
        this.parent();
    }
});

const ApplicationsButton = new Lang.Class({
    Name: 'ApplicationsButton',
    Extends: PanelMenu.Button,

    _init: function() {
        this.parent(1.0, null, false);

        this.setMenu(new ApplicationsMenu(this.actor, 1.0, St.Side.TOP, this));
        Main.panel.menuManager.addMenu(this.menu);

        // At this moment applications menu is not keyboard navigable at
        // all (so not accessible), so it doesn't make sense to set as
        // role ATK_ROLE_MENU like other elements of the panel.
        this.actor.accessible_role = Atk.Role.LABEL;

        let hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });

        this._label = new St.Label({ text: _("Applications"),
                                     y_expand: true,
                                     y_align: Clutter.ActorAlign.CENTER });
        hbox.add_child(this._label);
        hbox.add_child(new St.Label({ text: '\u25BE',
                                      y_expand: true,
                                      y_align: Clutter.ActorAlign.CENTER }));

        this.actor.add_actor(hbox);
        this.actor.name = 'panelApplications';
        this.actor.label_actor = this._label;

        this.actor.connect('captured-event', Lang.bind(this, this._onCapturedEvent));

        _showingId = Main.overview.connect('showing', Lang.bind(this, function() {
            this.actor.add_accessible_state (Atk.StateType.CHECKED);
        }));
        _hidingId = Main.overview.connect('hiding', Lang.bind(this, function() {
            this.actor.remove_accessible_state (Atk.StateType.CHECKED);
        }));

        this.reloadFlag = false;
        this._createLayout();
        this._display();
        _installedChangedId = appSys.connect('installed-changed', Lang.bind(this, function() {
            if (this.menu.isOpen) {
                this._redisplay();
            } else {
                this.reloadFlag = true;
            }
        }));

        // Since the hot corner uses stage coordinates, Clutter won't
        // queue relayouts for us when the panel moves. Queue a relayout
        // when that happens.
        _panelBoxChangedId = Main.layoutManager.connect('panel-box-changed', Lang.bind(this, function() {
            container.queue_relayout();
        }));
    },

    get hotCorner() {
        return Main.layoutManager.hotCorners[Main.layoutManager.primaryIndex];
    },

    _createVertSeparator: function() {
        let separator = new St.DrawingArea({ style_class: 'calendar-vertical-separator',
                                             pseudo_class: 'highlighted' });
        separator.connect('repaint', Lang.bind(this, this._onVertSepRepaint));
        return separator;
    },

    _onCapturedEvent: function(actor, event) {
        if (event.type() == Clutter.EventType.BUTTON_PRESS) {
            if (!Main.overview.shouldToggleByCornerOrButton())
                return true;
        }
        return false;
    },

    _onMenuKeyPress: function(actor, event) {
        let symbol = event.get_key_symbol();
        if (symbol == Clutter.KEY_Left || symbol == Clutter.KEY_Right) {
            let direction = symbol == Clutter.KEY_Left ? Gtk.DirectionType.LEFT
                                                       : Gtk.DirectionType.RIGHT;
            if (this.menu.actor.navigate_focus(global.stage.key_focus, direction, false))
                return true;
        }
        return this.parent(actor, event);
    },

    _onVertSepRepaint: function(area) {
        let cr = area.get_context();
        let themeNode = area.get_theme_node();
        let [width, height] = area.get_surface_size();
        let stippleColor = themeNode.get_color('-stipple-color');
        let stippleWidth = themeNode.get_length('-stipple-width');
        let x = Math.floor(width/2) + 0.5;
        cr.moveTo(x, 0);
        cr.lineTo(x, height);
        Clutter.cairo_set_source_color(cr, stippleColor);
        cr.setDash([1, 3], 1); // Hard-code for now
        cr.setLineWidth(stippleWidth);
        cr.stroke();
    },

    _onOpenStateChanged: function(menu, open) {
       if (open) {
           if (this.reloadFlag) {
               this._redisplay();
               this.reloadFlag = false;
           }
       }
       this.parent(menu, open);
    },

    _redisplay: function() {
        this.applicationsBox.destroy_all_children();
        this.categoriesBox.destroy_all_children();
        this._display();
    },

    _loadCategory: function(categoryId, dir) {
        let iter = dir.iter();
        let nextType;
        while ((nextType = iter.next()) != GMenu.TreeItemType.INVALID) {
            if (nextType == GMenu.TreeItemType.ENTRY) {
                let entry = iter.get_entry();
                if (!entry.get_app_info().get_nodisplay()) {
                    let app = appSys.lookup_app_by_tree_entry(entry);
                    let menu_id = dir.get_menu_id();
                    this.applicationsByCategory[categoryId].push(app);
                }
            } else if (nextType == GMenu.TreeItemType.DIRECTORY) {
                let subdir = iter.get_directory();
                if (!subdir.get_is_nodisplay())
                    this._loadCategory(categoryId, subdir);
            }
        }
    },

    scrollToPage: function(newPage) {
        // Get sizes
        let appsScrollBoxAdj = this.appsScrollBox.get_hscroll_bar().get_adjustment();
        let appsScrollBoxAlloc = this.appsScrollBox.get_allocation_box();
        let currentScrollValue = appsScrollBoxAdj.get_value();
        let boxWidth = appsScrollBoxAlloc.x2 - appsScrollBoxAlloc.x1;
        let pageSize = (APPLICATION_BOX_WIDTH * 4);
        let newScrollValue = newPage * pageSize;
        // Move scroll
        if (newScrollValue != currentScrollValue) {
            let timeline = new Clutter.Timeline ({ 'duration' : 200 });
            let distance = newScrollValue - currentScrollValue;
            let step = distance / timeline.duration;
            timeline.connect('new-frame', Lang.bind(this, function (timeline, msecs) {
                this.scrollToButtonAnimation (timeline, msecs, currentScrollValue, newScrollValue, step);
            } ));
            timeline.start();
        }
        // Select page number
        let actors = this.appsPagesBox.get_children();
        for (let i = 0; i < actors.length; i++) {
            let actor = actors[i];
            if (i==newPage) actor.set_style("background-color: #333;font-weight:1000;font-color: #fff;border-radius:5px;");
            else actor.set_style("font-weight:normal;");
        }
    },

    scrollToButton: function(button) {
        let buttonAlloc = button.actor.get_allocation_box();
        let pageSize = (APPLICATION_BOX_WIDTH * 4);
        let buttonPage = parseInt(buttonAlloc.x1 / pageSize);
        this.scrollToPage(buttonPage);
    },

    scrollToButtonAnimation: function(timeline, msecs, currentScrollValue, newScrollValue, step) {
        let position = currentScrollValue + step * msecs;
        let appsScrollBoxAdj = this.appsScrollBox.get_hscroll_bar().get_adjustment().set_value(position);
    },

    _buildFilterMenu: function() {

        let categoryAll = new CategoryMenuItem(this, "*");
        categoryAll.actor.set_accessible_name("*");
        this.filterMenu.menu.addMenuItem(categoryAll);

        this.applicationsByCategory = {};
        let tree = appSys.get_tree();
        let root = tree.get_root_directory();
        this.applicationsByCategory["*"] = [];
        let iter = root.iter();
        let nextType;
        while ((nextType = iter.next()) != GMenu.TreeItemType.INVALID) {
            if (nextType == GMenu.TreeItemType.DIRECTORY) {
                let dir = iter.get_directory();
                if (!dir.get_is_nodisplay()) {
                    let categoryId = dir.get_menu_id();
                    if (categoryId!="System Settings") { // "system settings" must be accessed from the application
                        this.applicationsByCategory[categoryId] = [];
                        this._loadCategory("*", dir);
                        this._loadCategory(categoryId, dir);
                        if (this.applicationsByCategory[categoryId].length > 0) {
                            let categoryItem = new CategoryMenuItem(this, dir);
                            categoryItem.actor.set_accessible_name(categoryId);
                            this.filterMenu.menu.addMenuItem(categoryItem);
                        }
                    }
                }
            }
        }
    },

    _createLayout: function() {
        this.filterMenu = new PopupMenu.PopupSubMenuMenuItem(_("Applications"));;
        this.filterMenu.status.text = _("All");
        this._buildFilterMenu ();
        this.filterMenu.menu.connect('open-state-changed', Lang.bind(this, function(obj, open) {
            if (open) {
                this.categorySelected = false;
                this.listOutBox.hide ();
                Tweener.addTween(this.listOutBox, { height:   0, time: 0.0 });
                Tweener.addTween(this.listOutBox, { opacity:  0, time: 0.0 });
            } else {
                if (!this.categorySelected) {
                        this.selectCategory(this.lastCategory);
                }
            }
        }));
        this.menu.addMenuItem(this.filterMenu);

        this.minWidthBox = new PopupMenu.PopupMenuSection();
        this.menu.addMenuItem(this.minWidthBox);
        this.minWidthLayout = new St.BoxLayout({ vertical: true });
        this.minWidthLayout.style=('height: 1px;width: '+(APPLICATION_BOX_WIDTH * 4)+'px;');
        this.minWidthBox.actor.add_actor(this.minWidthLayout);

        this.appsBox = new PopupMenu.PopupMenuSection();
        this.menu.addMenuItem(this.appsBox);

        this.listOutBox = new St.BoxLayout({ vertical: true });
        this.listOutBox.style=('height: 500px;width: '+(APPLICATION_BOX_WIDTH * 4)+'px;');
        this.appsBox.actor.add_actor(this.listOutBox);

        this.appsPagesBox = new St.BoxLayout({ vertical: false });
        this.listOutBox.add(this.appsPagesBox, { x_fill: false, y_fill: true});

        this.appsScrollBox = new St.ScrollView({ x_fill: true, y_fill: false,
                                                 x_align: St.Align.START,
                                                 style_class: 'luna_apps' });
        this.appsScrollBox.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.NEVER);
        let hscroll = this.appsScrollBox.get_hscroll_bar();
        hscroll.hide ();
        hscroll.connect('scroll-start', Lang.bind(this, function() {
            this.menu.passEvents = true; // TODO mirar com bloqueja els passEvents a la llista original!
        }));
        hscroll.connect('scroll-stop', Lang.bind(this, function() {
            this.menu.passEvents = false;
        }));
        this.appsListBox = new St.BoxLayout();
        this.appsScrollBox.add_actor(this.appsListBox);
        this.listOutBox.add(this.appsScrollBox, { expand: true, x_fill: true, y_fill: true });
    },

    _display: function() {
    },

    _clearApplicationsBox: function() {
        let actors = this.appsListBox.get_children();
        for (let i = 0; i < actors.length; i++) this.appsListBox.remove_actor(actors[i]);

        this.applicationsTable = new St.Table({ homogeneous: true, reactive: true });
        this.appsListBox.add_actor(this.applicationsTable);

        let pactors = this.appsPagesBox.get_children();
        for (let pi = 0; pi < pactors.length; pi++) this.appsPagesBox.remove_actor(pactors[pi]);
    },

    selectCategory: function(category) {
        this.lastCategory = category;
        this.categorySelected = true;

        let categoryLabel = _("All");
        if (category != "*")
            categoryLabel = category.get_name();

        this.filterMenu.status.text = categoryLabel;
        this.filterMenu.menu.close (true);
        this._clearApplicationsBox();

        let categoryName = "";
        if (category == "*")
            categoryName = category;
        else
            categoryName = category.get_menu_id();

        this._displayButtons(this._listApplications(categoryName));
    },

    _displayButtons: function(apps) {
        if (apps) {
            let page = 0;
            let posx = 0;
            let posy = 0;
            let listed = [];
            for (let i = 0; i < apps.length; i++) {
                let app = apps[i];
                let appName = app.get_name();
                if (listed.indexOf(appName) == -1) {
                    listed.push(appName);
                    let applicationMenuItem = new ApplicationMenuItem(this, app);
                    this.applicationsTable.add(applicationMenuItem.actor, { row: posy, col: (page * 4) + posx, x_fill: true, y_fill: true, 
                                                                            x_expand: true, y_expand: true,
                                                                            x_align: St.Align.MIDDLE, y_align: St.Align.MIDDLE});
                    posx++;
                    if (posx >= 4) {
                        posx = 0;
                        posy++;
                        if (posy >= 4) {
                            posy = 0;
                            page++;
                        }
                    }
                }
            }
            // Fill empty spaces to mantain proportions
            if (posx > 0) {
                for (let lx = posx; lx < 4; lx++) {
                    let lastBox = new St.Bin();
                    this.applicationsTable.add(lastBox, { row: posy, col: (page * 4) + lx, x_fill: true, y_fill: true, 
                                                          x_expand: true, y_expand: true,
                                                          x_align: St.Align.MIDDLE, y_align: St.Align.MIDDLE});
                }
            }
            // Define table size
            let pages = page+1;
            let tableStyle = ("width:"+((pages) * 4 * APPLICATION_BOX_WIDTH)+"px");
            this.applicationsTable.style = tableStyle;
            this.appsListBox.stye = tableStyle;

            // Add pages buttons
            if (pages > 1) {
                for (let px = 1; px <= pages; px++) {
                        let pageItem = new ApplicationPageItem(this, px);
                        this.appsPagesBox.add_child(pageItem.actor);
                }
            }

            // Show table
            let height = 500;
            if (pages == 1) {
                height = 25 + (posy + 1) * APPLICATION_BOX_WIDTH;
                if (posx == 0) height -= APPLICATION_BOX_WIDTH;
            }
            this.listOutBox.show ();
            Tweener.addTween(this.listOutBox, { height:  height, time: 0.15, transition: 'linear' });
            Tweener.addTween(this.listOutBox, { opacity: 255, time: 0.15, delay: 0.10, transition: 'linear' });

            // Set default page
            this.scrollToPage(0);
        }
    },

    _listApplications: function(category_menu_id) {
        let applist;

         if (category_menu_id) {
            applist = this.applicationsByCategory[category_menu_id];
            applist.sort(function(a,b) {
                return a.get_name().toLowerCase() > b.get_name().toLowerCase();
            });
        } 

        return applist;
    },

    destroy: function() {
        this.menu.actor.get_children().forEach(function(c) { c.destroy() });
        this.parent();
    }
});

let appsMenuButton;
let _hidingId;
let _installedChangedId;
let _panelBoxChangedId;
let _showingId;

function enable() {
    appsMenuButton = new ApplicationsButton();
    Main.panel.addToStatusArea('luna_apps', appsMenuButton, 1, 'left');

    Main.wm.setCustomKeybindingHandler('panel-main-menu',
                                       Shell.KeyBindingMode.NORMAL |
                                       Shell.KeyBindingMode.OVERVIEW,
                                       function() {
                                           appsMenuButton.menu.toggle();
                                       });
}

function disable() {
    Main.panel.menuManager.removeMenu(appsMenuButton.menu);
    appSys.disconnect(_installedChangedId);
    Main.layoutManager.disconnect(_panelBoxChangedId);
    Main.overview.disconnect(_hidingId);
    Main.overview.disconnect(_showingId);
    appsMenuButton.destroy();

    Main.wm.setCustomKeybindingHandler('panel-main-menu',
                                       Shell.KeyBindingMode.NORMAL |
                                       Shell.KeyBindingMode.OVERVIEW,
                                       Main.sessionMode.hasOverview ?
                                       Lang.bind(Main.overview, Main.overview.toggle) :
                                       null);
}

function init(metadata) {
    Convenience.initTranslations();
}
